//
//  MAIN PROGRAM FOR ENCODING
//
#include "defines.h"
#include "IO.h"

results psnrBitsRes;
int configurationFlag = 0;

//
//
//
void postProcessing(Bitstream *bitstream, imgpel **imgY_rec, imgpel **imgY_org, int frNo, int qp, int img_height, int img_width)
{
  static int first=0;
  static imgpel **imgY_var;
  int filtNo_best, varIndTab[NO_VAR_BINS], usePrevFiltBest[NO_VAR_BINS], prevFiltInd, i, j;
  double errorEnc;
  static double **filterCoeffSym;
  int force0;
  int codedFlag;

#if DEBUGGING==1
  double errorDec;
#endif

  if(first == 0)
  {
    initMatrix_ushort(&imgY_var, img_height, img_width); 
    initMatrix_double(&filterCoeffSym, NO_VAR_BINS, MAX_SQR_FILT_LENGTH); 
    first = 1;
  }
  // zero all variables - debugging purposes
  filtNo_best = 0; 
  prevFiltInd = 0;
  for(i = 0; i < NO_VAR_BINS; i++)
  {
    varIndTab[i] = 0;
    usePrevFiltBest[i] = 0;
    for(j = 0; j < MAX_SQR_FILT_LENGTH; j++)
      filterCoeffSym[i][j] = 0;
  }

  calcVar(imgY_var, imgY_rec, 3, img_height, img_width);

  errorEnc = findBestFilterVarPred(bitstream, imgY_rec, imgY_org, imgY_var, frNo, qp, &filtNo_best, filterCoeffSym, varIndTab, 
    usePrevFiltBest, &prevFiltInd, &force0, &codedFlag, img_height, img_width);

#if DEBUGGING==1
  errorDec = decFilterVarPred(imgY_rec, imgY_org, imgY_var, filtNo_best, filterCoeffSym, varIndTab, usePrevFiltBest, 
    prevFiltInd, codedFlag, img_height, img_width);
  assert(errorEnc == errorDec);
#endif
}

//
//
//
void printUsage(char *pname)
{
  printf("Usage:\n");
  printf("\t%s -w <width> -h <height> [-f <first frame>] [-s <skip>] -n <# frames> -q <QP> [-c <configuration>] -o <original yuv> -r <decoded yuv> [-t <output yuv>] [-b <bitstream>] [-x <coding order>]\n\n", pname);
  printf("Note:     Parameter <configuration> is equal to 0 for IPPP and 1 for Hierarchical B\n");
  printf("          The console output of the encoder can be used to specify coding order\n\n");
  printf("Defaults: <first frame> = 0\n");
  printf("          <skip> = 0\n");
  printf("          <configuration> = 0\n");
  printf("          <output> = Not saved\n");
  printf("          <bitstream> = Not saved\n");
  printf("          <coding order> = Coding and display order are the same\n\n");
  printf("Example:\n");
  printf("\t%s -w 1280 -h 720 -f 0 -s 0 -n 150 -q 32 -c 0 -o Raven.yuv -r Raven_dec.yuv -t Raven_pf.yuv -b bitstream.out -x CodingOrder.txt\n\n", pname);
  exit(-1);
}

//
//  Read a file containing lines starting with a zero padded, four digits number
//  Lines non starting with a digit are ignored
//  If two lines containing 0000 are read, the second one is skipped
//  The (redirected) console output of the encoder can be used as input
//
int readCodingOrderFile(int *frameNo, int numFrames, FILE * fp)
{
  char lineBuf[1024] = {0};
  int skipSecondZero = 0;
  int pos = -1;
  int numFramesRead = 0;
  int frame;
  
  while(!feof(fp))
  {
    fgets(lineBuf, 1023, fp);
    pos = strcspn(lineBuf, "0123456789");
    if(!pos)
    {
      if(sscanf(lineBuf, "%04d", &frame))
      {
        if(frame)
        {
          frameNo[numFramesRead++] = frame;
        }
        else
        {
          if(!skipSecondZero)
          {          
            frameNo[numFramesRead++] = frame;
            skipSecondZero = 1;
          }
        }
      }
    }
  }
  return numFramesRead;
}

//
//
//
int main(int argc, char *argv[])
{     
  int width = 0, height = 0, skip = 0, start = 0, imgNo = 0, noFr = 0, status = 0;
  FILE *p_orig, *p_rec, *p_out, *p_co, *p_tmp;
  char seqOrig[2048] = "";
  char seqRec[2048] = "";
  char seqTmp[2048] = "";
  char cOrder[2048] = NO_OUTPUT_FILE;
  char seqOut[2048] = NO_OUTPUT_FILE;
  char bsOut[2048] = NO_OUTPUT_FILE;
  imgStruct imgOrig, imgRec;
  int qp = 0;
  int imgStartSubSeq = 0;
  int *codingOrder;
  
  Bitstream *bitstream = allocBitstream(100000);

  if(argc < 13)
    printUsage("pfencod");

  while((argc > 1 ) && (argv[1][0] == '-'))
  {
    switch (argv[1][1])
    {
      case 'w':
        width = atoi(argv[2]);              break;
      case 'h':
        height = atoi(argv[2]);             break;
      case 's':
        skip = atoi(argv[2]);               break;
      case 'f':
        start = atoi(argv[2]);              break;
      case 'c':
        configurationFlag = atoi(argv[2]);  break;
      case 'o':
        strcpy(seqOrig, argv[2]);           break;
      case 'r':
        strcpy(seqRec, argv[2]);            break;
      case 't':
        strcpy(seqOut, argv[2]);            break;
      case 'b':
        strcpy(bsOut, argv[2]);             break;
      case 'x':
        strcpy(cOrder, argv[2]);            break;
      case 'n':
        noFr = atoi(argv[2]);               break;
      case 'q':
        qp = atoi(argv[2]);                 break;
      default:
        FATAL_ERROR_1("Bad option %s\n", argv[1], -1);
    }
    argc -= 2;
    argv += 2;
  }

  if(width == 0)
    FATAL_ERROR_1("Bad option %s\n", "width", -1);
  if(height == 0)
    FATAL_ERROR_1("Bad option %s\n", "height", -1);
  if(noFr == 0)
    FATAL_ERROR_1("Bad option %s\n", "# frames", -1);

  initImg(&imgOrig, height, width);       // Original frame (only used for psnr)
  initImg(&imgRec, height, width);        // Reconstructed

  if((p_orig = fopen(seqOrig, "rb")) == 0)
    FATAL_ERROR_1("Error open file seqOrig %s\n", seqOrig, -1);
  if((p_rec = fopen(seqRec, "rb")) == 0)
    FATAL_ERROR_1("Error open file seqRec %s\n", seqRec, -1);

  sprintf(seqTmp, "%s_%d_%d_%s", seqOrig, configurationFlag, qp, TMP_YUV_FILE);
  if((p_tmp = fopen(seqTmp, "wb")) == 0)
    FATAL_ERROR_1("Error open file seqTmp %s\n", seqTmp, -1);


  // Read coding order file
  codingOrder = (int*) calloc(noFr, sizeof(int));
  if(!strcmp(cOrder, NO_OUTPUT_FILE))
  {
    // Default
    for(imgNo = 0; imgNo < noFr; imgNo++)
      codingOrder[imgNo] = imgNo;
  }
  else
  {
    int noFrTmp = 0;
    if(!(p_co = fopen(cOrder, "r")))
      FATAL_ERROR_1("Error open file cOrder %s\n", cOrder, -1);
    noFrTmp = readCodingOrderFile(codingOrder, noFr, p_co);
    if(noFrTmp != noFr)
      FATAL_ERROR_2("Error: read %d frame numbers out of %d\n", noFrTmp, noFr, -1);
    fclose(p_co);
  }

  resultsInit(&psnrBitsRes);
  imgStartSubSeq = 0;
  imgNo = 0;

  do {
    status = read_one_new_frame(height, width, 0, 0, codingOrder[imgNo], p_rec, &imgRec);
    status += read_one_new_frame(height, width, skip, start, codingOrder[imgNo], p_orig, &imgOrig);
    if(status == 0)
    {
      postProcessing(bitstream, imgRec.imgY, imgOrig.imgY, imgNo, qp, height, width);
      write_image(height, width, p_tmp, imgRec);        // Tmp output in coding order
      find_snr(&psnrBitsRes, imgOrig.imgY, imgRec.imgY, height, width);
      printResults(&psnrBitsRes, codingOrder[imgNo]);
      imgNo++; 
    }
  } while((status == 0) && (imgNo < noFr));
  
  fclose(p_tmp);
  fclose(p_rec);
  fclose(p_orig);

  // Sort output from coding to display order
  if(!strcmp(cOrder, NO_OUTPUT_FILE))
  {
    if(!strcmp(seqOut, NO_OUTPUT_FILE))
    {
      if(remove(seqTmp) == -1)
        FATAL_ERROR_1("Error deleting %s\n", seqTmp, -1);
    }
    else
    {
      if(rename(seqTmp, seqOut))
        FATAL_ERROR_1("Error renaming %s\n", seqTmp, -1);
    }
  }
  else
  {
    int *frPos = (int*) calloc(noFr, sizeof(int));

    if((p_tmp = fopen(seqTmp, "rb")) == 0)
      FATAL_ERROR_1("Error open file seqTmp %s\n", seqTmp, -1);

    if((p_out = fopen(seqOut, "wb")) == 0)
      FATAL_ERROR_1("Error open file seqOut %s\n", seqOut, -1);

    for(imgNo = 0; imgNo < noFr; imgNo++)
      frPos[codingOrder[imgNo]] = imgNo;

    for(imgNo = 0; imgNo < noFr; imgNo++)
      if(!read_one_new_frame(height, width, 0, 0, frPos[imgNo], p_tmp, &imgRec))
        write_image(height, width, p_out, imgRec);

    fclose(p_out);
    fclose(p_tmp);
    free(frPos);

    // Delete temporary output
    if(remove(seqTmp) == -1)
      FATAL_ERROR_1("Error deleting %s\n", seqTmp, -1);
  }

  // Report
  printResultsFinal(&psnrBitsRes);
  resultsEnd(psnrBitsRes);

  // Output bitstream
  saveBitstream(bitstream, bsOut);
  freeBitstream(bitstream);

  return (0);
}

